from quple.components.descriptors import *
from quple.circuits.qubit_register import QubitRegister
from quple.circuits.quantum_circuit import QuantumCircuit
from quple.circuits.templates.template_circuit_block import TemplateCircuitBlock
from quple.circuits.parameterised_circuit import ParameterisedCircuit
from quple.circuits.templates.pauli_block import PauliBlock
from quple.circuits import variational_circuits
from quple.circuits.common_circuits import *
from quple.utils.utils import *

MAX_WORKERS = 8
CV_NJOBS = -1


from typing import List
import re
import copy
import cirq
import quple
import numpy as np
import sympy as sp
import numba
from concurrent.futures  import ProcessPoolExecutor


def replace_symbol_in_op(op, old_symbol:sp.Symbol, new_symbol:sp.Symbol) -> None:
    """Replace symbols in a parameterised gate operation with new symbols
    
    Args:
        op: gate operation to which the associated symbols are replaced
        old_symbol: the original symbol associated with the gate operation
        new_symbol: the new symbol which the original symbol is to be replaced by
    """
    if isinstance(op, cirq.EigenGate):
        if old_symbol in op.exponent.free_symbols:
            op._exponent = op.exponent.subs(old_symbol, new_symbol)

    if isinstance(op, cirq.FSimGate):
        if isinstance(op.theta, sympy.Basic):
            if old_symbol in op.theta.free_symbols:
                op._theta = op.theta.subs(old_symbol, new_symbol)
        if isinstance(op.phi, sympy.Basic):
            if old_symbol in op.phi.free_symbols:
                op._phi = op.phi.subs(old_symbol, new_symbol)

    if isinstance(op, cirq.PhasedXPowGate):
        if isinstance(op.exponent, sympy.Basic):
            if old_symbol in op.exponent.free_symbols:
                op._exponent = op.exponent.subs(old_symbol, new_symbol)
        if isinstance(op.phase_exponent, sympy.Basic):
            if old_symbol in op.phase_exponent.free_symbols:
                op._phase_exponent = op.phase_exponent.subs(old_symbol, new_symbol)
                
def resolve_expression_map_conflicts(old_expr_map, new_expr_map):
    """Recovers reflattened expressions from a new expression map given the old expression map
    
    Args:
        old_expr_map: cirq.ExpressionMap
            The original expression map which maps flattened expressions to unflattened expressions
        new_expr_map: cirq.ExpressionMap
            The new expression map which is possibly reflattened from a flattened expressions
    Returns:
        An expression map with reflattened expressions recovered
    """    
    inverted_map = {v:k for k,v in old_expr_map.items()}
    reflattened = set(inverted_map).intersection(set(new_expr_map.values()))
    expr_map = cirq.ExpressionMap({(k if v not in reflattened else inverted_map[v]):v 
                                   for k,v in new_expr_map.items()})
    return expr_map
                
def pqc_symbol_map(circuit:cirq.Circuit, symbols_map) -> cirq.Circuit:
    """Maps the old symbols in a circuit with the new ones
    
    Args:
        circuit: cirq.Circuit, quple.QuantumCircuit
            The circuit to map the parameter symbols
        symbols_map:
            A dictionary that maps old parameter symbols to new ones
            
    Returns:
        A new circuit with the old parameter symbols replaced by the new ones
    """
    new_circuit = copy.deepcopy(circuit)
    for moment in new_circuit:
        for op in moment:
            if cirq.is_parameterized(op):
                symbols_in_op = quple.symbols_in_op(op.gate)
                for sym in symbols_in_op:
                    replace_symbol_in_op(op.gate, sym, symbols_map[sym])
    return new_circuit



def merge_pqc(circuits:List[cirq.Circuit], symbol:str='θ') -> cirq.Circuit:
    """Merges a list of parameterized circuit and updates the parameter symbols
    
    Circuits are merged in the order they are listed. The set of symbols in one circuit 
    will be treated as distinct from the set of symbols in another circuit. The merged
    circuit will have all parameter symbols of the form {symbol}_{index}. 
    
    Args:
        circuits: list of cirq.Circuit or quple.QuantumCircuit
            The circuits to merge
        symbol: str
            The parameter symbol prefix for the merged circuit

    Returns:
        The merged circuit
    """        
    symbol_size = 0
    if not all(isinstance(circuit, cirq.Circuit) for circuit in circuits):
        raise ValueError('Circuits to be merged must be intances of cirq.Circuit object')
    circuits = [circuit.get_unflattened_circuit() if isinstance(circuit, quple.QuantumCircuit) \
                else circuit for circuit in circuits]
    for circuit in circuits:
        symbol_size += len(quple.get_circuit_symbols_in_order(circuit))
    all_symbols = sp.symarray(symbol, symbol_size)
    qubits = set()
    for circuit in circuits:
        qubits |= set(quple.get_circuit_qubits(circuit))
    qubits = sorted(list(qubits))
    merged_circuit = quple.QuantumCircuit(qubits)
    for circuit in circuits:
        old_symbols = quple.get_circuit_symbols_in_order(circuit)
        new_symbols = all_symbols[:len(old_symbols)]
        all_symbols = all_symbols[len(old_symbols):]
        symbols_map = {old:new for old, new in zip(old_symbols, new_symbols)}
        merged_circuit.append(pqc_symbol_map(circuit, symbols_map))
    return merged_circuit

def plot_stacked_roc_curve(fpr_list:List[np.ndarray], 
                   tpr_list:List[np.ndarray], labels:List[str]=None, title:str='ROC Curve', with_auc:bool=True):
    """Plots multiple roc curves stacked together
    
    Args:
        fpr: list/array of numpy array
            A collection of arrays containing the false positive rates for 
            different experiments
        tpr: list/array of numpy array
            A collection of arrays containing the false positive rates for 
            different experiments
        labels: list of str, default = None
            List of labels for different experiments.
        title: str, default = 'ROC Curve'
            Title of the plot
        with_auc: True or False
            Whether to include the auc in the labels

    Returns:
        A matplotlib plot of the stacked roc curve
    """     
    assert len(fpr_list) == len(tpr_list)
    n = len(fpr_list)
    if labels is None:
        labels = ['']*n
    assert len(labels) == n
    from sklearn.metrics import roc_curve, auc
    import matplotlib.pyplot as plt
    plt.clf()
    plt.rcParams["font.weight"] = "bold"
    plt.xlabel("Signal Efficiency", fontsize=18,fontweight='bold')
    plt.ylabel("Background Rejection", fontsize=18,fontweight='bold')
    plt.title(title, fontsize=16,fontweight='bold')
    plt.xlim(0.0, 1.0)
    plt.ylim(0.0, 1.0)
    plt.grid(color='gray', linestyle='--', linewidth=1)  
    for fpr,tpr,label in zip(fpr_list, tpr_list, labels):
        if with_auc:
            roc_auc = auc(fpr, tpr)
            label += ', AUC={:.4f}'.format(roc_auc)
        plt.plot(tpr,1-fpr, linestyle='-',label=label,linewidth=2)
    plt.plot([0, 1], [1, 0], linestyle='--', color='black', label='Luck, AUC= 0.5')
    plt.legend(loc='best',prop={'size': 8})    
    return plt       


def plot_roc_curve(fpr:np.ndarray, tpr:np.ndarray, label:str='', title:str='ROC Curve', with_auc:bool=True):
    """Plots a roc curve
    
    Args:
        fpr: numpy array
            An array containing the false positive rates
        tpr: numpy array
            An array containing the true positive rates
        label: str
            Label of the curve
        title: str
            Title of the plot
        with_auc: True or False
            Whether to include the auc in the labels
    
    Returns:
        A matplotlib plot of the roc curve
    """        
    return plot_stacked_roc_curve([fpr], [tpr], [label], title=title, with_auc=with_auc)    

def atoi(symbol):
    return int(symbol) if symbol.isdigit() else symbol

def natural_key(symbol):
    '''Keys for human sorting
    Reference:
    http://nedbatchelder.com/blog/200712/human_sorting.html
    '''
    return [ atoi(s) for s in re.split(r'(\d+)', symbol.name) ]

def parallel_run(fuc, *iterables, max_workers=None):
    n_iteration = len(iterables)
    if n_iteration == 0:
        return []
    elif n_iteration == 1:
        return [func(iterables[0], *iterables)]

    max_workers = max_workers or quple.MAX_WORKERS 

    with ProcessPoolExecutor(max_workers) as executor:
        result = executor.map(fuc, *iterables)
    return [i for i in result]


def batching(l:List, n:int):
    for i in range(0, len(l), n):
        yield l[i:i + n]
    
def flatten_list(l):
    return [item for sublist in l for item in sublist]


from typing import Optional, Sequence, List, Dict, Union
import numpy as np
import numba


from sklearn import svm
from sklearn.model_selection import GridSearchCV

import qiskit
import quple
from .qsvm_logger import QSVMLogger

from quple.utils.mathext import abs2, DataPrecision

from pdb import set_trace

class DefaultDict(dict):
    def __missing__(self, key):
        return key


class QSVM:
    __DEFAULT_SVC_PARAM_GRID__ = {
        'C': [1,2,3,4,5,8,9,10,15,20,30,50,100,200,400,800,1000],
    }
    
    def __init__(self, encoding_circuit:Union[quple.ParameterisedCircuit, qiskit.QuantumCircuit], 
                 precision:Union[str,DataPrecision]='double', **kwargs):
        self._attributes = DefaultDict({})
        self.encoding_circuit = encoding_circuit
        self.precision = precision
        if self.circuit_type == 'qiskit':
            if 'backend' not in kwargs:
                raise ValueError('Backend must be given for a qiskit circuit')
            else:
                self.backend = kwargs['backend']
    @property
    def precision(self):
        return self._precision
    
    @precision.setter
    def precision(self, val):
        if isinstance(val, str):
            self._precision = DataPrecision[val]
        elif isinstance(val, DataPrecision):
            self._precision = val
        else:
            raise Value('Invalid type for numeric precision: {}'.format(type(val)))
                
    
    @property
    def entangle_strategy(self):
        return self._attributes['entangle_strategy']
    
    @property
    def attributes(self):
        return self._attributes
    
    @property
    def backend(self):
        return self._backend
    
    @backend.setter
    def backend(self, val):
        if self._precision == DataPrecision['single']:
            val.backend.set_options(precision='single')
        elif self._precision == DataPrecision['double']:
            val.backend.set_options(precision='double')
        else:
            pass
            #raise ValueError('Invalid value for numerical precision: {}'.format(self._precision))
        self._backend = val
    
    @property
    def n_qubit(self):
        return self._attributes['n_qubit']
    @property
    def encoding_map(self):
        return self._attributes['encoding_map']
    
    @property
    def circuit_type(self):
        return self._attributes['circuit_type']
    
    @property
    def circuit_name(self):
        return self._attributes['circuit_name']    
        
    @property
    def encoding_circuit(self):
        return self._encoding_circuit
  
    @encoding_circuit.setter
    def encoding_circuit(self, val):
        if isinstance(val, quple.ParameterisedCircuit):
            self._attributes['circuit_type'] = 'quple'
            self._attributes['n_qubit'] = val.n_qubit
            self._attributes['encoding_map'] = val.encoding_map.__name__ if \
                isinstance(val, quple.data_encoding.EncodingCircuit) else '' 
            self._attributes['entangle_strategy'] = val.entangle_strategy if \
                isinstance(val, quple.ParameterisedCircuit) and \
                isinstance(val.entangle_strategy, str) else ''
            self._attributes['circuit_name'] = val.name
            self._encoding_circuit = val
        elif isinstance(val, qiskit.QuantumCircuit):
            self._attributes['circuit_type'] = 'qiskit'
            self._attributes['n_qubit'] = val.num_qubits
            self._attributes['encoding_map'] = val._data_map_func.__name__ 
            self._attributes['entangle_strategy'] = val._entanglement if \
                isinstance(val, qiskit.circuit.library.NLocal) else '' 
            self._attributes['circuit_name'] = type(val).__name__
            self._encoding_circuit = val
        elif isinstance(val, qiskit.aqua.components.feature_maps.FeatureMap):
            self._attributes['circuit_type'] = 'qiskit'
            self._attributes['n_qubit'] = val._num_qubits
            self._attributes['encoding_map'] = ''
            self._attributes['entangle_strategy'] = ''
            self._attributes['circuit_name'] = type(val).__name__
            self._encoding_circuit = val
        else:
            raise ValueError('Circuit type currently not supported: {}'.format(type(encoding_circuit))) 
    
    @staticmethod
    def get_kernel_matrix(state_vectors_left, state_vectors_right):
        return abs2(state_vectors_left @ state_vectors_right.conjugate().T)

    def get_state_vectors(self, x):
        if self.circuit_type == 'quple':
            return self.encoding_circuit.get_state_vectors(x)
        elif self.circuit_type == 'qiskit':
            from quple.qiskit_interface.tools import get_qiskit_state_vectors
            return get_qiskit_state_vectors(self.backend, self.encoding_circuit, x)
        else:
            return None
    
    @staticmethod 
    def train_svc(train_kernel_matrix, y_train, random_seed=0,
                  param_grid=None, n_jobs=-1, cv=3, scoring='roc_auc', **kwargs):
        param_grid = param_grid or QSVM.__DEFAULT_SVC_PARAM_GRID__
        
        svc = svm.SVC(kernel='precomputed', probability=True, random_state=random_seed)
        clf = GridSearchCV(estimator=svc, param_grid=param_grid, n_jobs=n_jobs,
                           cv=cv, scoring=scoring)

        clf.fit(train_kernel_matrix, y_train)
        return clf
    
    def run_with_tune(self, x_train, y_train, x_val, y_val, 
                      x_test, y_test, random_seed = 0, logger=None, **kwargs):
                             
        # setup logger
        if logger:
            if isinstance(logger, QSVMLogger):
                attributes = {}
                attributes['feature_dimension'] = x_train.shape[1]
                attributes['train_size'] = x_train.shape[0]
                attributes['val_size'] = x_val.shape[0]
                attributes['test_size'] = x_test.shape[0]
                attributes.update(self.attributes)
                logger.set_attributes(attributes)
            else:
                raise ValueError('Only quple.QSVMLogger is supported')

        if logger:
            logger.on_train_start()

        if logger:
            logger.info('Evaluating state vectors for train data...')             
        train_state_vectors = self.get_state_vectors(x_train)
        if logger:
            logger.info('Evaluating state vectors for validation data...')             
        val_state_vectors = self.get_state_vectors(x_val)
        if logger:
            logger.info('Evaluating state vectors for test data...')             
        test_state_vectors = self.get_state_vectors(x_test)
        if logger:
            logger.info('Evaluating kernel matrix for train data...')
        train_kernel_matrix = self.get_kernel_matrix(train_state_vectors, train_state_vectors)
        if logger:
            logger.info('Evaluating kernel matrix for validation data...')
        val_kernel_matrix = self.get_kernel_matrix(val_state_vectors, train_state_vectors)
        if logger:
            logger.info('Evaluating kernel matrix for test data...')
        test_kernel_matrix = self.get_kernel_matrix(test_state_vectors, train_state_vectors)
        
        train_state_vectors = None # Free memory
        val_kernel_matrix = None # Free memory
        test_state_vectors = None  # Free memory 

        clf = self.train_svc(train_kernel_matrix, y_train, random_seed=random_seed)    
    

        logger.on_train_end(clf, train_kernel_matrix, test_kernel_matrix, y_test, val_kernel_matrix, y_val)
        return clf
    
    def run(self, x_train, y_train, x_test, y_test, 
            random_seed = 0, logger=None, **kwargs):
                             
        # setup logger
        if logger:
            if isinstance(logger, QSVMLogger):
                attributes = {}
                attributes['feature_dimension'] = x_train.shape[1]
                attributes['train_size'] = x_train.shape[0]
                attributes['test_size'] = x_test.shape[0]
                attributes.update(self.attributes)
                logger.set_attributes(attributes)
            else:
                raise ValueError('Only quple.QSVMLogger is supported')
        if logger:
            logger.on_train_start()
        
        if logger:
            logger.info('Evaluating state vectors for train data...')             
        train_state_vectors = self.get_state_vectors(x_train)
        if logger:
            logger.info('Evaluating kernel matrix for train data...')
        train_kernel_matrix = self.get_kernel_matrix(train_state_vectors, train_state_vectors) 
        if logger:
            logger.info('Fitting SVM model...')
        clf = self.train_svc(train_kernel_matrix, y_train, random_seed=random_seed, n_jobs=quple.CV_NJOBS)    
        if logger:
            logger.info('Evaluating state vectors for test data...')             
        test_state_vectors = self.get_state_vectors(x_test)        
        if logger:
            logger.info('Evaluating kernel matrix for test data...')
        test_kernel_matrix = self.get_kernel_matrix(test_state_vectors, train_state_vectors)
        
        train_state_vectors = None # Free memory
        test_state_vectors = None  # Free memory 

        logger.on_train_end(clf, train_kernel_matrix, test_kernel_matrix, y_test)
        return clf


from enum import Enum
import psutil
import multiprocessing

class ByteExpression(Enum):
    KB = 1
    MB = 2
    GB = 3
    TB = 4
    PB = 5
    
def get_cpu_count():
    return multiprocessing.cpu_count()

def get_gpu_count():
    import torch
    return torch.cuda.device_count()

def get_gpu_device_names():
    import torch
    return [torch.cuda.get_device_name(i) for i in range(get_gpu_count)]

def get_virtual_memory(byte_expr:ByteExpression = ByteExpression.GB):
    return psutil.virtual_memory().total/1024**byte_expr.value

def get_available_virtual_memory(byte_expr:ByteExpression = ByteExpression.GB):
    return psutil.virtual_memory().available/1024**byte_expr.value

def get_used_virtual_memory(byte_expr:ByteExpression = ByteExpression.GB):
    return psutil.virtual_memory().used/1024**byte_expr.value


from enum import Enum
import numba

class DataPrecision(Enum):
    single = 1
    double = 2

@numba.vectorize([numba.float64(numba.complex128),numba.float32(numba.complex64)])
def abs2(x):
    return x.real**2 + x.imag**2